import javax.microedition.io.*;
import java.io.*;
import java.util.*;

import com.ajile.drivers.gpio.*;
import com.ajile.components.*;
import com.ajile.events.*;

public class SecurityWebserver {

    // Default HTTP port is 80
    private final int myListenPort = 80;

    // Store the last MAX_EVENTS events
    private final int MAX_EVENTS = 10;
    private Vector events = new Vector(MAX_EVENTS);

    private String header = "HTTP/1.0 200 Document follows\r\n" +
       "Content-Type: text/html\r\n" +
       "Connection: close\r\n" +
       "\r\n" +
       "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\n" +
       "<HTML><HEAD><TITLE>Security Event Log</TITLE><HEAD>";
    
    public SecurityWebserver() {
        // aJile examples suggest using this class instead of 
        // StreamConnectionNotifier
        com.sun.cldc.io.j2me.serversocket.Protocol scn = null;
        initializeGpio();
        try {
            scn = (com.sun.cldc.io.j2me.serversocket.Protocol)Connector.open("serversocket://:" + myListenPort, Connector.READ_WRITE);
        } catch (IOException ioe) {
            System.out.println("Can't open server socket: "+ioe.getMessage());
        }
        while (true) {
            try {
                StreamConnection sc = scn.acceptAndOpen();
                InputStream is = sc.openInputStream();
                OutputStream os = sc.openOutputStream();
                
                eatInputStream(is);

                StringBuffer buf = new StringBuffer(header);
                buf.append("<BODY><TABLE border=\"1\"><TR><TD><B>Sensor Number</B><TD><B>How Long Ago</B>");
                long now = System.currentTimeMillis();
                for (int i=events.size()-1; i>=0; i--) {
                    SecurityEvent event = (SecurityEvent)events.elementAt(i);
                    buf.append("<TR><TD align=center>");
                    buf.append(String.valueOf(event.detectorNumber));
                    buf.append("<TD>");
                    buf.append(event.getTime(now));
                }
                buf.append("</TABLE></BODY></HTML>");
                buf.append("\r\n\r\n");
                os.write(buf.toString().getBytes());
                os.close();
                sc.close();
            } catch (IOException ioe) {
                System.err.println("IOException: "+ioe.getMessage());
            }
        }
    }
    
    public static void main(String [] args) {
        new SecurityWebserver();
    }
    
    /**
     * Read and discard the input stream until the end of the HTTP header
     */
    private void eatInputStream(InputStream is) throws IOException {
        int [] next = {-1, 0, 0, 0};
        int [] pat = {'\r', '\n', '\r', '\n'};
        int pIdx = -1;
        int ch = 0;
        while (true) {
            if (ch < 0) {
                System.out.println("End Of Stream shouldn't happen!");
                break; // end of stream?
            }
            if ((pIdx == -1) ||  (ch == pat[pIdx])) {
                pIdx++;
                if (pIdx == pat.length) break; // match!
                ch = is.read();
            } else {
                pIdx = next[pIdx];
            }
        }
        is.close();
    }
    
   
    /*
     * A little class to store the information about the security event
     */
    private static class SecurityEvent {
        int detectorNumber;
        long time;
        final int millisPerSec = 1000;
        final int millisPerMin = millisPerSec*60;
        final int millisPerHour = millisPerMin*60;
        public SecurityEvent(int detectorNumber, long time) {
            this.detectorNumber = detectorNumber;
            this.time = time;
        }
        String getTime(long now) {
            long duration = now - time;
            long hours = duration / millisPerHour;
            duration -= hours * millisPerHour;
            long mins = duration / millisPerMin;
            duration -= mins * millisPerMin;
            long secs = duration / millisPerSec;
            return ""+hours+" hours, "+mins+" minutes, "+secs+" seconds";
        }
    }
    
   
    /**
     * Initialize the GPIO pins and register listeners
     */
    private void initializeGpio() {
        int [] pins = {
            GpioPin.GPIOE_BIT0, GpioPin.GPIOE_BIT1,
            GpioPin.GPIOE_BIT2, GpioPin.GPIOE_BIT3,
            GpioPin.GPIOE_BIT4, GpioPin.GPIOE_BIT5,
            GpioPin.GPIOE_BIT6, GpioPin.GPIOE_BIT7};
            
            for (int i=0; i<pins.length; i++) {
                PushButton inPin = new PushButton(pins[i]);
                inPin.setTriggerReportPolicy(PushButton.TRIGGER_ON_PUSH);
                inPin.addTriggerListener(new EventListener(i));
            }
    }
    
    /**
     * Inner class is a listener for events from a PushButton object
     */
    private class EventListener implements TriggerEventListener {
        int detectorNumber;
        public EventListener(int detectorNumber) {
            this.detectorNumber = detectorNumber;
        }
        public void triggerEvent() {
            SecurityEvent me = new SecurityEvent(detectorNumber, 
                                             System.currentTimeMillis());
            synchronized (events) {
                if (events.size() == MAX_EVENTS)
                    events.removeElementAt(0);
                events.addElement(me);
            }
        }
    }
    

}
