open Misc2 SmlMC_head TextIO Polyhash Process

val \ = fn(a,c) => Array.sub(a,c)
infix 9 \
val \\ = fn(a,c) => Vector.sub(a,c)
infix 9 \\

val update=Array.update
val array=Array.array

val events = dut.events

val time = ref 0

fun for(a,b,F) = if a>b then () else (F(a); for(a+1,b,F))

fun pr_state st =
  for(0,(Vector.length st)-1,fn i => dut.print_comp(st\\i))

exception HashExn
val sta_table = mkTable ((hash_safe_param 500 1000),op=) (2000,HashExn)
val tra_table = mkTable ((hash_safe_param 20 50),op=) (4000,HashExn)
val sta_array = Dynarray.array(2000,[])

val sta_cnt = ref 0
val sta_inc_cnt = ref 0

type sta_Value = { pos:int,
                   fst_time:int,
                   prec:int list ref,
                   out:int list ref,
                   num:int ref,
                   pred:int ref }

fun sta_lookup (state,ev_nr,tq,dift:int) =
  let
  val key = (state,ev_nr,tq)
  val time = !time+dift
  val value:sta_Value =
    { pos= !sta_cnt, fst_time=time, prec=ref[], out=ref[], num=ref 0, pred=ref 0 } 
  val key_value = (key,value)
  in
    case peekInsert sta_table key_value of
      NONE => (value,false) before
      ( if not(!quiet) then
        ( pr.printq `  sta: ^(pr.d(!sta_cnt)) - `; pr_state state;
          if !verbose then pr.printq ` ev:(^(pr.lis ev_nr (fn i=>pr.d i)))`
          else pr.printq ` ev:^(pr.d (List.length ev_nr))`;
          let
            val len = List.length tq
            fun pr (te:timed_events) =
              pr.sprintq `^(pr.d(#delta te))(^(pr.lis(#te_active te)(fn i=>pr.d i)))`
          in
            if len>0 then 
              if !verbose then pr.printq ` tq:^(pr.lis tq pr)`
              else pr.printq ` tq:^(pr.d len)`
            else ()
          end;
          if time>0 then pr.printq ` time:^(pr.d time)` else ();
          print "\n"
        ) else ();
        Dynarray.update(sta_array,!sta_cnt,[key_value]);
        incr sta_cnt;
        incr sta_inc_cnt;
        if !sta_inc_cnt=1000 then
        ( sta_inc_cnt := 0;
          pr.printq `^(pr.d(!sta_cnt)) states^("\n")` ;
          flushOut stdOut
        ) else ()
      )
    | SOME v => (v,true)
  end

fun tra_lookup (orig_pos,pos,dift) =
( case peekInsert tra_table ((orig_pos,pos,dift),()) of
    NONE => false before
    ( if not(!quiet) then
      ( if dift > 0 then
          pr.printq `  tra: ^(pr.d orig_pos) ^(pr.d pos) dt:^(pr.d dift)^("\n")`
        else
          pr.printq `  tra: ^(pr.d orig_pos) ^(pr.d pos)^("\n")`
      ) else ()
    )
  | SOME _ => true
)

val new_states = ref []

fun collect_evnr n = 
( if n = !events_len then []
  else if !(#active (events\\n)) then
  ( #active (events\\n) := false;
    n::collect_evnr(n+1)
  )
  else collect_evnr(n+1)
)

val comps_len = Vector.length dut.components

fun copy_comps() =
  Vector.tabulate(comps_len,fn i => !(dut.components\\i))

fun update_new_st orig_pos orig_tq dift =
  let
  val cmp=copy_comps()
  val ev_nr=collect_evnr 0
  val (v,sta_known)=sta_lookup(cmp,ev_nr,orig_tq,dift)
  val tra_known = tra_lookup(orig_pos,#pos v,dift)
  in
    if not sta_known then
      new_states := (#pos v,#fst_time v,cmp,ev_nr,orig_tq) :: !new_states
    else ();
    if not tra_known then
      #prec v := !(#prec v) @ [orig_pos]
    else ()
  end

val ev_stack = ref[]

fun run_state (orig_st,orig_pos,mut_vec,evs,ev_max,dift) =
  let
  val rec expand_ev = fn
    ([],n,old_n) =>
    ( for (0,(Vector.length orig_st)-1,fn i => dut.components\\i := orig_st\\i);
      iter
        ( fn e =>
          ( if !verbose then print "fire!\n" else ();
            e();  (* execute an event *)
            if not dut.with_timing then
            ( if dut.model_check then (!path_test)() else ();
              update_new_st orig_pos (!timing_queue) dift
            )
            else ()
          )
        )
        ev_stack;
      if dut.with_timing then 
      ( if dut.model_check then (!path_test)() else ();
        update_new_st orig_pos (!timing_queue) dift
      )
      else ()
    )
  | (ev::evs,n,old_n) =>
    let
      val rec exp_list = fn
        ([],_,_) => ()
      | (e::es,i,old_n) =>
        ( push e ev_stack;
          expand_ev(evs,i+1,old_n);
          ignore(pop ev_stack);
          exp_list(es,i,old_n)
        )
    in
      case #ev(events\\ev) of
        Ev e =>
        ( push e ev_stack;
          expand_ev(evs,n+1,old_n);
          ignore(pop ev_stack)
        )
      | (Ev_list es) =>
        exp_list(es,n,n)
    end
  val rec mut = fn
    n =>
    if n < ev_max then
      evs\\(mut_vec\n)::mut(n+1)
    else []
  in
    ev_stack := [];
    expand_ev (mut 0,0,0)
  end

fun visit_all (nr,k,orig_st,orig_pos,evs,ev_max,mut_vec,dift) =
( update(mut_vec,k,nr);
  let val nr=nr+1
  in
    if nr=ev_max then
      run_state(orig_st,orig_pos,mut_vec,evs,ev_max,dift)
    else ();
    for(0,ev_max-1,fn i =>
      if mut_vec\i = ~1 then visit_all(nr,i,orig_st,orig_pos,evs,ev_max,mut_vec,dift)
      else ()
    )
  end;
  update(mut_vec,k,~1)
)

val fac_permmax = 
  let val rec f = fn 1 => 1 | n => n * f(n-1)
  in
    if !perm_max=0 then 0 else f (!perm_max)
  end

val random_gen = Random.newgen()

fun visit_rand (orig_st,orig_pos,evs,ev_max,dift) =
  let
  val selmax =
    let val rec f = fn
      1 => 1
    | n => if n<=ev_max then n * f(n-1) else fac_permmax
    in
      f (!perm_max)
    end
  val used = array(ev_max,false)
  val mut_vec = array(ev_max,~1)
  in
    for (0,selmax-1,fn _ =>
    ( for(0,ev_max-1,fn i => update(used,i,false));
      for(0,ev_max-1,
        fn i => 
        ( let val rval=ref(Random.range(0,ev_max-1) random_gen)
          in
            while used\(!rval) do rval := Random.range(0,ev_max) random_gen;
            update(used,!rval,true);
            update(mut_vec,i,!rval)
          end
        )
      );
      run_state(orig_st,orig_pos,mut_vec,evs,ev_max,dift)
    )
  )
  end

fun dyna_sub(a,i) = hd(Dynarray.sub(a,i))

fun dyna_app f a stop =
  let
  fun lr j = if j < stop then (f(dyna_sub(a,j)); lr(j+1))
             else ()
  in lr 0
  end

fun
  visit [] _ = ()
| visit ((_,_,_,[],[])::sts) dift =
    visit sts dift
| visit ((orig_pos,orig_time,orig_st,[],tq_hd::tq_tl)::sts) dift =
  ( timing_queue := tq_tl;
    visit
      ((orig_pos,orig_time,orig_st,#te_active tq_hd,tq_tl)::sts)
      (#delta tq_hd)
  )
| visit ((orig_pos,orig_time,orig_st,orig_ev,tq)::sts) dift =
  let
  val ev_max=List.length orig_ev
  in
    cur_state_nr := orig_pos;
    time := orig_time;
    if !perm_max = 0 orelse ev_max <= !perm_max then
    ( let val mut_vec = array(ev_max,~1)
      in
        visit_all(~1,0,orig_st,orig_pos,Vector.fromList orig_ev,ev_max,mut_vec,dift)
      end
    )
    else
      visit_rand(orig_st,orig_pos,Vector.fromList orig_ev,ev_max,dift);
    if !verbose then
      pr.printq `new states=^(pr.d(length(!new_states)))^("\n")`
    else ();
    visit sts dift
  end

fun restore (comps, ev_nr:int list, tq:timed_events list) =
  for (0,comps_len-1,fn i => dut.components\\i := comps\\i)

val cc_cnt = ref 1
val cc_stack = ref[]

(* From: Data structures and algorithms in C++ (Adam Drozdek):
 * strongDFS(v)
 *   pred(v)=num(v)=i++;
 *   push(v);
 *   for <all vertices> u <adjacent to> v
 *     if num(u) == 0
 *       strongDFS(u)
 *       pred(v)=min(pred(v),pred(u))
 *     else if num(u) < num(v) && u <on stack>
 *       pred(v)=min(pred(v),num(u))
 *   if pred(v)==num(v)
 *     w=pop()
 *     while w!=v
 *       <output> w
 *       w=pop()
 *     <output> w
*)
fun strongDFS nr =
  let
  fun stval nr =
    let val (_,v:sta_Value) = dyna_sub(sta_array,nr)
    in v
    end
  fun stkey nr =
    let val (st,_) = dyna_sub(sta_array,nr)
    in st
    end
  val v = stval nr
  fun min a b = if !a > !b then a := !b else ()
  in
    #pred v := !cc_cnt;
    #num v := !cc_cnt;
    incr cc_cnt;
    push nr cc_stack;
    iter
    ( fn out =>
      let
      val u = stval out
      in
        if !(#num u) = 0 then
        ( strongDFS out;
          min (#pred v) (#pred u)
        )
        else if !(#num u) < !(#num v) andalso (on_stack out (op=) cc_stack) then
          min (#pred v) (#num u)
        else ()
      end
    ) (#out v);
    if !(#pred v) = !(#num v) then 
      let
      val nr2 = ref(pop cc_stack)
      val (cc_buf,cc_outg) = (ref[],ref[])
      in
        if !nr2 = nr then
        ( if null (!(#out v)) then
          ( if not(!quiet) then pr.printq `  deadlock: ^(pr.d nr)^("\n")` else ();
            restore (stkey nr);
            cur_state_nr := nr;
            (!reach_test) true
          )
          else if on_stack nr (op=) (#out v) then
          ( if not(!quiet) then pr.printq `  final: ^(pr.d nr)^("\n")` else ();
            restore (stkey nr);
            cur_state_nr := nr;
            (!reach_test) true
          )
          else ()
        )
        else
        ( push (!nr2) cc_buf;
          while !nr2 <> nr do
          ( let val outg = #out(stval(!nr2)) in
              if length (!outg) > 1 then
                iter (fn out => push out cc_outg) outg
              else ()
            end;
            nr2 := pop cc_stack;
            push (!nr2) cc_buf
          );
          (* now cc_outg contains all outgoing edges, cc_buf all strongly connected states *)

          let
          val out_edge = ref false
          val init = ref true
          in
            iter
            ( fn n =>
              if not(on_stack n (op=) cc_buf) then out_edge := true
              else ()
            ) cc_outg;

            if not(!out_edge) then
            ( if not(!quiet) then
              ( print "  final: ";
                iter
                ( fn n => pr.printq `^(pr.d n) `
                ) cc_buf;
                print "\n"
              ) else ();
              iter
              ( fn n => 
                ( restore (stkey n);
                  cur_state_nr := n;
                  if !init then 
                  ( init := false;
                    (!reach_test) true
                  )
                  else (!reach_test) false
                )
              ) cc_buf
            ) else ()
          end
        )
      end
    else ()
  end

fun eval_states() =
  let
  in
    dyna_app  (* assign #out of predecessor *)
    ( fn (_,v) =>
      iter
      ( fn pre =>
        let
        val (_,prec) = dyna_sub(sta_array,pre)
        in
          #out prec := (#pos v) :: !(#out prec)
        end
      ) (#prec v)
    ) sta_array (!sta_cnt);

    strongDFS 0
  end

val rec generate = fn
  [] =>
  ( if dut.model_check then eval_states() else ();
    pr.printq `^(pr.d(!sta_cnt)) states^("\n")`
  )
| sts =>
  ( new_states := [];
    visit sts 0;
    generate(!new_states)
  )

fun set_global_vars () =
  let
  fun set_option x =
  ( case x of
      "-verb" => verbose := true
    | "-q"    => quiet := true
    | "-h"    =>
      ( print "Runtime options:\n";
        print "   -verb   Verbose output\n";
        print "   -q      Quiet mode (no sta, tra, final output)\n";
        exit failure
      )
    | _       => err("Unkown option: "^x^"\n")
  )
  fun search_options [] = ()
    | search_options (x::xs) =
      ( set_option x;
        search_options xs
      )
  in
    search_options(CommandLine.arguments())
  end

fun main() =
( let
  val cmp=copy_comps()
  val e=collect_evnr 0
  in
    backtrace_ref :=
    ( fn nr =>
      let
      val cur_state = ref(dyna_sub(sta_array,nr))
      fun get_v (_,v) = v
      in
        print "  backtrace:";
        while #pos(get_v(!cur_state)) <> 0 do
        ( pr.printq ` ^(pr.d(#pos(get_v(!cur_state))))`;
          cur_state := dyna_sub(sta_array,hd(!(#prec(get_v(!cur_state)))))
        );
        print " 0\n"
      end
    );
    set_global_vars();
    if !events_len = 0 then err "no events" else ();
    print "start: "; pr_state cmp; print "\n";
    flushOut stdOut;
    ignore (sta_lookup (cmp,e,[],0));
    generate([(0,0,cmp,e,[])]);
    exit success
  end
)

val _ = main() handle Fail s => (print "error: "; print s; print "\n")
