structure SmlMC_head = struct
local open Misc2 TextIO in

val verbose = ref false
val quiet = ref false

datatype event =
  Ev of (unit->unit)
| Ev_list of (unit->unit) list

type timed_events = { delta:int, te_active:int list }
type event_record={ active:bool ref, index:int, ev:event }

fun err s = raise Fail s
val events_len = ref 0

fun ev (e:event_record ref) f =
( e := { active=ref false, index= !events_len, ev=Ev f };
  incr events_len;
  !e
)

fun ev_choice (e:event_record ref) fs =
( e := { active=ref false, index= !events_len, ev=Ev_list fs };
  incr events_len;
  !e
)

fun nop() = ref { active=ref false, index=0, ev=Ev(fn()=>err "nop event") }
fun act (e:event_record ref) = #active(!e) := true

val perm_max = ref 0
val ignore_events = ref false (* in hashtable lookup *)
val timing_queue = ref([]:timed_events list)

fun init_ev (evs:event_record list) =
  Vector.tabulate(!events_len, fn i =>
    let val e = List.nth(evs,i)
    in
      if #index e <> i then pr.err `init_ev: index mismatch ^(pr.d(#index e)),^(pr.d i)`
      else ();
      e
    end
  )

fun act_timed (ev:event_record ref) t =
  let
  val ev_index = #index(!ev)
  fun newte td = { delta=td, te_active=[ev_index] }
  val rec update_tim = fn
    ([],_) => []
  | (te::tes,tdiff) =>
    { delta=(#delta te)-tdiff, te_active=(#te_active te) }::update_tim(tes,tdiff)
  val rec insert = fn
    ([],td) => [newte td]
  | (te::tes,td) =>
    if td=(#delta te) then
    ( let
        fun fnd _ [] = false
          | fnd i (n::ns) = if i=n then true else fnd i ns
        fun ins i [] = [i]
          | ins i (n::ns) = if i<n then i::n::ns else n::ins i ns
      in
        if fnd ev_index (#te_active te) then te::tes
        else { delta=td, te_active=ins ev_index (#te_active te) }::tes
      end
    )
    else if td < (#delta te) then
      (newte td)::update_tim (te::tes,td)
    else
      te::insert(tes,td-(#delta te))
  in
    timing_queue := insert(!timing_queue,t)
  end

val reach_test = ref(fn(init:bool) => ())
val path_test = ref(fn() => ())
val cur_state_nr = ref 0
val backtrace_ref = ref(fn (n:int) => ())
fun backtrace() = (!backtrace_ref)(!cur_state_nr)
end
end
