(* Demonstrating data access with a local structure
   Also try with: with_timing=true
*)

structure dut = struct
local
  open SmlMC_head
in
val with_timing = false
val model_check = false

datatype Component =
  Buffer of int*int  (* cnt, max *)

structure buf = struct
  fun get(ref(Buffer d)) = d
  fun set(b,f) = b := Buffer(f(get b))
  fun cnt b = #1(get b)
  fun max b = #2(get b)
end

val buffer = ref(Buffer(0,3))
val components = #[ buffer ]
val (e1,e2) = (nop(),nop())

val events = init_ev
[ ev e1
  ( fn() =>
    ( let val n = buf.cnt buffer
      in
        if n < buf.max buffer then buf.set(buffer,fn(_,m)=>(n+1,m)) else ()
      end;
      act e1
    )
  ),
  ev e2
  ( fn() =>
    ( let val n = buf.cnt buffer
      in
        if n > 0 then buf.set(buffer,fn(_,m)=>(n-1,m)) else ()
      end;
      act e2
    )
  )
]

val _ = ( act e1; act e2 )

val print_comp = fn
  Buffer(n,_) => pr.printq `buffer(^(pr.d n))`

end
end 
