(* Demonstrating timed events
   Filling the buffer is faster then emptying - also try the opposite
   Also run with option -verb (verbose output: event- and timing-queue listed)
*)

structure dut = struct
local
  open SmlMC_head
in
val with_timing = true
val model_check = false

datatype Component =
  Buffer of int*int  (* cnt, max *)

structure buf = struct
  fun get(ref(Buffer d)) = d
  fun set(b,f) = b := Buffer(f(get b))
  fun cnt b = #1(get b)
  fun max b = #2(get b)
end

val buffer = ref(Buffer(0,3))
val components = #[ buffer ]
val (e1,e2) = (nop(),nop())

fun mk_ev e up t =
  fn() =>
  ( let val n = buf.cnt buffer
    in
      if up andalso n < buf.max buffer then buf.set(buffer,fn(_,m)=>(n+1,m))
      else if not up andalso n > 0 then buf.set(buffer,fn(_,m)=>(n-1,m))
      else ()
    end;
    act_timed e t
  )

val events = init_ev
[ ev e1 (mk_ev e1 true 2),
  ev e2 (mk_ev e2 false 3)
]

val _ = ( act e1; act e2 )

val print_comp = fn
  Buffer(n,_) => pr.printq `buffer(^(pr.d n))`

end
end 
