structure dut = struct
local
  open SmlMC_head
in
 
val with_timing = true
val model_check = true

datatype State = one|two|relay
datatype Component =
  Node of int * int * State * int list  (* d, e, state, out *)
| Global of bool * int                  (* leader_known, nr_leaders *)

structure node = struct
  fun get(ref(Node d)) = d | get _ = err "get(node)"
  fun d dat     = #1 dat
  fun e dat     = #2 dat
  fun state dat = #3 dat
  fun out dat   = #4 dat
  fun set nod f = nod := Node(f(get nod))
end

structure global = struct
  fun get(ref(Global d)) = d | get _ = err "get(global)"
  fun set glob f = glob := Global(f(get glob))
end

val (n0,n1,n2,n3,n4,glob) =
( ref(Node(3,0,one,[3])),
  ref(Node(2,0,one,[2])),
  ref(Node(1,0,one,[1])),
  ref(Node(5,0,one,[5])),
  ref(Node(4,0,one,[4])),
  ref(Global(false,0))
)

val components = #[n0,n1,n2,n3,n4,glob]

val (ev_n0,ev_n1,ev_n2,ev_n3,ev_n4) = (nop(),nop(),nop(),nop(),nop())

fun node(pre,id,ev_id) = (fn() =>
  let
    val nod = node.get id
    val pred = node.get pre
  in
    case (node.state nod) of
    one =>
    ( if null(node.out pred) then ()
      else
        let val e=hd (node.out pred)
        in
          if e=(node.d nod) then
          ( if e=5 then () else pr.printq `leader = ^(pr.d e)`;
            node.set id (fn(v1,_,_,v4)=>(v1,e,relay,v4 @ [node.d nod]));
            global.set glob (fn(v1,v2)=>(v1,v2 + 1))
          )
          else
          ( node.set id (fn(v1,_,_,v4)=>(v1,e,two,v4 @ [e]))
          );
          node.set pre (fn(v1,v2,v3,v4)=>(v1,v2,v3,tl v4))
        end;
      act ev_id
    )
    | two =>
    ( if null(node.out pred) then ()
      else
        let val f=hd(node.out pred)
        in
          if (node.e nod) >= Int.max(f,node.d nod)
          then node.set id (fn(v1,v2,_,v4)=>(v2,v2,one,v4 @ [v1]))
          else node.set id (fn(v1,v2,_,v4)=>(v1,v2,relay,v4 @ [v1]));
          node.set pre (fn(v1,v2,v3,v4)=>(v1,v2,v3,tl v4))
        end;
      act ev_id
    )
    | relay =>
    ( if null(node.out pred) then ()
      else
      ( node.set id (fn(v1,v2,_,v4)=>(v1,v2,relay,v4 @ [hd(node.out pred)]));
        node.set pre (fn(v1,v2,v3,v4)=>(v1,v2,v3,tl v4))
      );
      act ev_id
    )
  end
)

val events = init_ev
[ ev ev_n0 (node(n4,n0,ev_n1)),
  ev ev_n1 (node(n0,n1,ev_n2)),
  ev ev_n2 (node(n1,n2,ev_n3)),
  ev ev_n3 (node(n2,n3,ev_n4)),
  ev ev_n4 (node(n3,n4,ev_n0))
]

val _ =
(
  act ev_n0;act ev_n1;act ev_n2;act ev_n3;act ev_n4;

  path_test := ( fn() =>
    let
      val (known,nr) = global.get glob
    in
      if not known andalso nr > 0 then
        global.set glob (fn(_,v2)=>(true,v2))
      else if known andalso nr <> 1 then
        pr.err `path: ^(pr.d nr) leaders, leader known`
      else ()
    end
  );

  reach_test := ( fn init =>
    let
      val (known,nr) = global.get glob
    in
      if nr <> 1 then
      ( backtrace();
        pr.err `final: ^(pr.d nr) leaders`
      )
      else ()
    end
  )
)

val print_comp = fn
  Node(d,e,state,out) =>
  let
    val s = fn one=>"one"|two=>"two"|relay=>"rel"
  in
    pr.printq `node(d=^(pr.d d) e=^(pr.d e) state=^(s state) out=[^(pr.lis out pr.d)]) `
  end
| Global(lkn,nrl) =>
    pr.printq `ld_kn=^(pr.b lkn) nr_ld=^(pr.d nrl) `

end
end
