structure dut = struct
local
  open SmlMC_head
in

val with_timing = true
val model_check = true

val \ = fn(a,n) => Vector.sub(a,n) handle _ => pr.err `subsc = ^(pr.d n)`
infix 9 \

fun nth(lis,n) = List.nth(lis,n) handle _ => pr.err `nth = ^(pr.d n)`

fun replace (n,lis,rep) =
  let
    val rec repl = fn
      (_,[]) => pr.err `repl: n=^(pr.d n)`
    | (m,x::xs) =>
      if n=m then (rep x)::xs
      else x::repl(m+1,xs)
  in
    repl(0,lis)
  end

val ( Asubsc,    (* originator *)
      Bsubsc,    (* responder *)
      NoTone,    (* tones *)
      DialT,
      BusyT,
      CongT,
      ErrorT,
      RingS,
      RingT,
      SpeechConn
    ) = (10,11,12,13,14,15,16,17,18,19);
val ( Onh,       (* hook status *)
      Offh,
      NoDigit,   (* illegal user codes *)
      ErrorDigit,
      BusyDigit,
      Dial,      (* dialing *)
      on,
      off,
      dc         (* dont care *)
    ) = (20,21,22,23,24,25,26,27,28)

val legal = fn
  0 => true
| 1 => true
| n => if n=NoDigit orelse n=ErrorDigit orelse n=BusyDigit
       then false
       else pr.err `legal: unexpected ^(pr.d n)`

datatype Component =
  Subs of int *    (* status: Onh/Offh/RingS/DialT/ErrorT *)
          int *    (* connected: on/off/Asubsc/Bsubsc *)
          string   (* state *)
| Mult of ( int *  (* id *)
            int *  (* called subscriber *)
            int *  (* on/off hook: Onh/Offh *)
            int    (* tone: NoTone/... *)
          ) list
| Next_state of
          int      (* next state nr *)

structure subs = struct
  val subs = #[ ref(Subs(Onh,off,"start")),
                ref(Subs(Onh,off,"start"))
              ]
  fun get n = case subs\n of ref(Subs d) => d | _ => err "get"
  fun get_data n = let val (a,b,_) = get n in (a,b) end
  fun status d = #1 d
  fun conn d =   #2 d
  fun set n f = subs\n := Subs(f(get n))
end

val next_state = #[ref(Next_state ~1),ref(Next_state ~1)]
val get_nxt = fn ref(Next_state m) => m | _ => err "get_nxt"

structure mult = struct
  val tset = ref(Mult[(0,NoDigit,Onh,NoTone),(1,NoDigit,Onh,NoTone)])
  fun get_tset() = case tset of ref(Mult t) => t | _ => err "get_tset"
  fun id d =     #1 d
  fun called d = #2 d
  fun hook d =   #3 d
  fun tone d =   #4 d

  val setAct = fn
    (n,onh,Mult ts) => Mult(replace(n,ts,fn(a,b,c,d)=>(a,b,onh,d)))
  | _ => err "setAct"

  val setBSubsc = fn
    (n,cald,Mult ts) => Mult(replace(n,ts,fn(a,b,c,d)=>(a,cald,c,d)))
  | _ => err "setBSubsc"

  val setTone = fn
    (n,ton,Mult ts) => Mult(replace(n,ts,fn(a,b,c,d)=>(a,b,c,ton)))
  | _ => err "setBSubsc"

  fun isBNumber n = (* called by other subscriber? *)
    let val rec isbn = fn
        [] => false
      | x::xs => if (called x)=n then true else isbn xs
    in isbn(get_tset())
    end

  fun isCalledBy n = (* called by other subscriber? *)
    let val rec isbn = fn
        (_,[]) => err "isCalledBy: unknown"
      | (m,x::xs) => if (called x)=n then (id x) else isbn(m+1,xs)
    in isbn(0,get_tset())
    end

  fun actStat n = nth(get_tset(),n)
  fun bSubsc n = called(nth(get_tset(),n))
end

val 
[ subscriber, callerOffh, calledRingS, calledOffh, 
  callerDialTOn, errorTone, errorToneOnh, calledOffhRingSOff, 
  speechConnCalled, speechConnCalledOnh, speechConnUserOff, callerDial,
  callerDialTOnOnh, callerDialTOff, busyTone, speechConnCaller,
  callerRingTOn, callerRingTOff, callerRingTOnOnh, busyToneOnh,
  speechConnCallerOnh,
  next_ev
] = List.tabulate(22,fn _ => #[nop(),nop()])

val components = #[ subs.subs\0, next_state\0, subs.subs\1, next_state\1, mult.tset ]

structure ok = struct
  local open mult
  in
  fun okay n =
  let
    val data = subs.get_data n
    val ts = actStat n
    fun == ((a1,b1),(a,b)) =
      (a=dc orelse a1=a) andalso
      (b=dc orelse b1=b)
    infix 5 ==
  in
    if data == (Offh,Asubsc) then
      if (legal n) andalso not(isBNumber n) andalso (hook ts)=Onh
      then ( tset := setAct(n,Offh,!tset); true )
      else false
    else if data == (Offh,Bsubsc) then
      if (legal n) andalso (isBNumber n) andalso (hook ts)=Onh
      then ( tset := setAct(n,Offh,!tset); true )
      else false
    else if data == (Onh,dc) then
      if legal(bSubsc n) andalso hook(actStat(bSubsc n))=Offh then
      ( tset := setBSubsc(
          bSubsc n,ErrorDigit,
          setBSubsc(n,NoDigit,setAct(n,Onh,!tset)));
        true
      )
      else if
        not(legal(bSubsc n))
        orelse
        legal(bSubsc n) andalso hook(actStat(bSubsc n))=Onh then
      ( tset := setBSubsc(n,NoDigit,setAct(n,Onh,!tset)); true )
      else false
    else if data == (Dial,dc) then
      let val cald = subs.conn data
      in
        if legal cald andalso n<>cald then
          if (tone ts)=DialT andalso 
            hook(actStat cald)=Onh andalso
            tone(actStat cald)=NoTone andalso
            ( not(isBNumber cald)
              orelse
              (isBNumber cald) andalso (bSubsc n)=cald
            )
          then ( tset := setBSubsc(n,cald,!tset); true )
          else if
            (tone ts)=DialT andalso (legal cald) andalso
            ( hook(actStat cald)=Offh orelse
              (isBNumber cald) andalso (bSubsc n)<>cald orelse
              tone(actStat cald)<>NoTone
            )
          then ( tset := setBSubsc(n,BusyDigit,!tset); true )
          else if
            (tone ts)=DialT andalso not(legal cald)
          then ( tset := setBSubsc(n,ErrorDigit,!tset); true )
          else if
            (tone ts)<>DialT
          then true
          else false
        else false
      end
    else if data == (DialT,on) then
      if (hook ts)=Offh andalso (tone ts)=NoTone andalso (bSubsc n)=NoDigit
      then ( tset := setTone(n,DialT,!tset); true )
      else false
    else if data == (DialT,off) then
      if (tone ts)=DialT
      then ( tset := setTone(n,NoTone,!tset); true )
      else false
    else if data == (RingS,on) then
      if (hook ts)=Onh andalso (isBNumber n) andalso
         (legal n) andalso (tone ts)=NoTone
      then ( tset := setTone(n,RingS,!tset); true )
      else false
    else if data == (RingS,off) then
      if (tone ts)=RingS andalso (legal n) andalso
         ( (hook ts)=Offh
           orelse
           not(isBNumber n)
         )
      then ( tset := setTone(n,NoTone,!tset); true )
      else false
    else if data == (RingT,on) then
      if (legal n) andalso (tone ts)=NoTone andalso
         legal(bSubsc n) andalso hook(actStat(bSubsc n))=Onh
      then ( tset := setTone(n,RingT,!tset); true )
      else false
    else if data == (RingT,off) then
      if (legal n) andalso
         ( (bSubsc n)<>NoDigit andalso hook(actStat(bSubsc n))=Offh
           orelse
           (bSubsc n)=NoDigit
         )
      then ( tset := setTone(n,NoTone,!tset); true )
      else false
    else if data == (ErrorT,on) then
      if (bSubsc n)=ErrorDigit andalso (tone ts)=NoTone
      then ( tset := setBSubsc(n,NoDigit,setTone(n,ErrorT,!tset)); true )
      else false
    else if data == (ErrorT,off) then
      if (tone ts)=ErrorT andalso (hook ts)=Onh
      then ( tset := setTone(n,NoTone,!tset); true )
      else false
    else if data == (BusyT,on) then
      if (bSubsc n)=BusyDigit andalso (tone ts)=NoTone
      then ( tset := setBSubsc(n,NoDigit,setTone(n,BusyT,!tset)); true )
      else false
    else if data == (BusyT,off) then
      if (tone ts)=BusyT andalso (hook ts)=Onh
      then ( tset := setTone(n,NoTone,!tset); true )
      else false
    else if data == (SpeechConn,on) then
      if (hook ts)=Offh andalso
         (tone ts)=NoTone andalso
         ( legal(bSubsc n) andalso
           hook(actStat(bSubsc n))=Offh
           orelse
           (isBNumber n) andalso
           hook(actStat(isCalledBy n))=Offh
         )
      then ( tset := setTone(n,SpeechConn,!tset); true )
      else false
    else if data == (SpeechConn,off) then
      if (tone ts)=SpeechConn andalso
         ( (hook ts)=Onh
           orelse
           (hook ts)=Offh andalso (bSubsc n)=ErrorDigit
         )
      then ( tset := setTone(n,NoTone,!tset); true )
      else if
        (tone ts)=SpeechConn andalso (hook ts)=Offh andalso
        legal(bSubsc n) andalso hook(actStat(bSubsc n))=Onh
      then ( tset := setBSubsc(n,ErrorDigit,setTone(n,NoTone,!tset)); true )
      else false
    else
      let val (a,b)=data
      in
        pr.err `okay (^(pr.d a),^(pr.d b))`
      end
  end
end
end

val events_ref = ref (#[]:event_record Vector.vector)

fun if_okay(nxt,n) =
  if ok.okay n then
  ( next_state\n := Next_state(~1); act(nxt\n) )
  else
  ( next_state\n := Next_state(#index(!(nxt\n))); act(next_ev\n) )

fun subscribers n =
  let
    val set = subs.set
  in
  [
  ev_choice (subscriber\n)
    [ fn() => (set n (fn _ =>(Offh,Asubsc,"1")); if_okay(callerOffh,n)),
      fn() => (set n (fn _ =>(RingS,on,"2")); if_okay(calledRingS,n))
    ],
  ev_choice (callerOffh\n)
    [ fn() => (set n (fn _ =>(DialT,on,"3")); if_okay(callerDialTOn,n)),
      fn() => (set n (fn _ =>(ErrorT,on,"4")); if_okay(errorTone,n))
    ],
  ev_choice (callerDialTOn\n)
    [ fn() => (set n (fn _ =>(Dial,(n+1)mod 2,"5")); if_okay(callerDial,n)),
      fn() => (set n (fn(_,b,_)=>(Onh,b,"6")); if_okay(callerDialTOnOnh,n))
    ],
  ev (callerDial\n)
    (fn() => (set n (fn _ =>(DialT,off,"7")); if_okay(callerDialTOff,n))),
  ev (callerDialTOnOnh\n)
    (fn() => (set n (fn _ =>(DialT,off,"8")); if_okay(subscriber,n))),
  ev_choice (callerDialTOff\n)
    [ fn() => (set n (fn(_,b,_)=>(Onh,b,"9")); if_okay(subscriber,n)),
      fn() => (set n (fn _ =>(RingT,on,"10")); if_okay(callerRingTOn,n)),
      fn() => (set n (fn _ =>(BusyT,on,"11")); if_okay(busyTone,n)),
      fn() => (set n (fn _ =>(ErrorT,on,"12")); if_okay(errorTone,n)),
      fn() => (set n (fn _ =>(SpeechConn,on,"13")); if_okay(speechConnCaller,n))
    ],
  ev_choice (callerRingTOn\n)
    [ fn() => (set n (fn(_,b,_)=>(Onh,b,"14")); if_okay(callerRingTOnOnh,n)),
      fn() => (set n (fn _ =>(RingT,off,"15")); if_okay(callerRingTOff,n))
    ],
  ev_choice (callerRingTOff\n)
    [ fn() => (set n (fn(_,b,_)=>(Onh,b,"16")); if_okay(subscriber,n)),
      fn() => (set n (fn _ =>(SpeechConn,on,"17")); if_okay(speechConnCaller,n))
    ],
  ev (callerRingTOnOnh\n)
    (fn() => (set n (fn _ =>(RingT,off,"18")); if_okay(subscriber,n))),
  ev (busyTone\n)
    (fn() => (set n (fn(_,b,_)=>(Onh,b,"19")); if_okay(busyToneOnh,n))),
  ev (busyToneOnh\n)
    (fn() => (set n (fn _ =>(BusyT,off,"20")); if_okay(subscriber,n))),
  ev_choice (speechConnCaller\n)
    [ fn() => (set n (fn(_,b,_)=>(Onh,b,"21")); if_okay(speechConnCallerOnh,n)),
      fn() => (set n (fn _ =>(SpeechConn,off,"22")); if_okay(speechConnUserOff,n))
    ],
  ev (speechConnCallerOnh\n)
    (fn() => (set n (fn _ =>(SpeechConn,off,"23")); if_okay(subscriber,n))),
  ev_choice (calledRingS\n)
    [ fn() => (set n (fn _ =>(Offh,Bsubsc,"24")); if_okay(calledOffh,n)),
      fn() => (set n (fn _ =>(RingS,off,"25")); if_okay(subscriber,n))
    ],
  ev (calledOffh\n)
    (fn() => (set n (fn _ =>(RingS,off,"26")); if_okay(calledOffhRingSOff,n))),
  ev_choice (calledOffhRingSOff\n)
    [ fn() => (set n (fn(_,b,_)=>(Onh,b,"27")); if_okay(subscriber,n)),
      fn() => (set n (fn _ =>(SpeechConn,on,"28")); if_okay(speechConnCalled,n))
    ],
  ev_choice (speechConnCalled\n)
    [ fn() => (set n (fn(_,b,_)=>(Onh,b,"29")); if_okay(speechConnCalledOnh,n)),
      fn() => (set n (fn _ =>(SpeechConn,off,"30")); if_okay(speechConnUserOff,n))
    ],
  ev (speechConnCalledOnh\n)
    (fn() => (set n (fn _ =>(SpeechConn,off,"31")); if_okay(subscriber,n))),
  ev_choice (speechConnUserOff\n)
    [ fn() => (set n (fn(_,b,_)=>(Onh,b,"32")); if_okay(subscriber,n)),
      fn() => (set n (fn _ =>(ErrorT,on,"33")); if_okay(errorTone,n))
    ],
  ev (errorTone\n)
    (fn() => (set n (fn(_,b,_)=>(Onh,b,"34")); if_okay(errorToneOnh,n))),
  ev (errorToneOnh\n)
    (fn() => (set n (fn _ =>(ErrorT,off,"35")); if_okay(subscriber,n)))
  ]
  end

fun nxt n =
[ ev (next_ev\n) (fn()=>
  ( if ok.okay n then
    ( act(ref((!events_ref)\(get_nxt(next_state\n)))); next_state\n := Next_state(~1)
    )
    else act(next_ev\n)
  ))
]

val events = init_ev
( (subscribers 0) @ (subscribers 1) @ (nxt 0) @ (nxt 1)
)

val _ =
( events_ref := events;
  act(subscriber\0); act(subscriber\1)
)

val print_comp = fn
  Subs(a,b,c) =>
  pr.printq `subs(status=^(pr.d(subs.status(a,b))) conn=^(pr.d(subs.conn(a,b))) state=^c) `
| Mult ds =>
  let
    fun pr d =
    pr.sprintq `(id=^(pr.d(mult.id d)) called=^(pr.d(mult.called d)) hook=^(
                pr.d(mult.hook d)) tone=^(pr.d(mult.tone d)))`
  in pr.printq `mult^(pr.lis ds pr) `
  end
| Next_state i => pr.printq `nxt(^(pr.d i)) `

end end
