structure dut = struct
local
  open SmlMC_head
in
val with_timing = true
val model_check = true

datatype State = RC | AWAKE | PREQ | LISTEN | ROOT | CHILD
val prs = fn RC=>"RC" | AWAKE=>"AWAKE" | PREQ=>"PREQ"
           | LISTEN=>"LISTEN" | ROOT=>"ROOT" | CHILD=>"CHILD"

datatype Component =
  Node of int*State*int*(int list) (* recb,state,del,obuf *)

val (nomesg,preq,ack)=(0,1,2)
fun prc c = Vector.sub(#["nomesg","preq","ack"],c)

val fast = 4
val slow = 7
val gamma = 2

structure node = struct
  val get = fn ref(Node t) => t
  fun set this f = this := Node(f(get this))
  fun recb   t = #1(get t) (* receive buffer *)
  fun state  t = #2(get t) (* FSM state *)
  fun del    t = #3(get t) (* delay RC -> AWAKE *)
  fun obuf   t = #4(get t) (* output buffer *)
end

val (n1,n2)=
( ref(Node(nomesg,RC,fast,[preq])),
  ref(Node(nomesg,RC,fast,[preq]))
)

val (ev_n1,w1,ev_n2,w3) = (nop(),nop(),nop(),nop())

val components = #[ n1,n2 ]

fun wire(snd,rcv,rcv_ev) =
  fn() =>
    if null(node.obuf snd) then ()
    else
    ( node.set rcv (fn(r,s,d,ob)=>(hd(node.obuf snd),s,d,ob));
      node.set snd (fn(r,s,d,ob)=>(r,s,d,tl ob));
      if node.state rcv = PREQ then act rcv_ev
      else ()
    )

fun node_ev(id,nod,del2,wire_ev) =
  fn() =>
  ( case node.state nod of
      RC =>
      ( act_timed id (node.del nod);
        node.set nod (fn(r,s,d,ob)=>(r,AWAKE,del2,ob))
      )
    | AWAKE =>
      let
        val recb = node.recb nod
      in
        if recb = ack then
          node.set nod (fn(r,s,d,ob)=>(r,CHILD,d,ob))
        else if recb = preq then
        ( node.set nod (fn(r,s,d,ob)=>(r,ROOT,d,ob @ [ack]));
          act_timed wire_ev gamma
        )
        else if recb = nomesg then 
        ( node.set nod (fn(r,s,d,ob)=>(r,PREQ,d,if null ob then [preq] else ob));
          act_timed wire_ev gamma
        )
        else ()
      end
    | PREQ =>
      ( if node.recb nod = nomesg then ()
        else
        ( node.set nod (fn(r,s,d,ob)=>(r,LISTEN,d,ob));
          act id
        )
      )
    | LISTEN =>
      let
        val recb = node.recb nod
      in
        if recb = ack then
          node.set nod (fn(r,s,d,ob)=>(r,CHILD,d,ob))
        else if recb = preq then
        ( node.set nod (fn(r,s,d,ob)=>(nomesg,RC,d,ob));
          act id
        )
        else pr.err `unexpected recb ^(pr.d recb)`
      end
    | _ => ()
  )

val events = init_ev
[ ev ev_n1 (node_ev(ev_n1,n1,fast,w1)),
  ev w1 (wire(n1,n2,ev_n2)),
  ev ev_n2 (node_ev(ev_n2,n2,slow,w3)),
  ev w3 (wire(n2,n1,ev_n1))
]

val _ =
( act ev_n1; act ev_n2;

  reach_test := (fn init =>
    let 
      val state_n1 = node.state n1
      val state_n2 = node.state n2
    in
      if state_n1=CHILD andalso state_n2=ROOT then ()
      else pr.err `^("\n")fast=^(pr.d fast), slow=^(pr.d slow), gamma=^(
        pr.d gamma), state ne1=^(prs state_n1), state ne2=^(prs state_n2)`
    end
  )
)

val print_comp = fn
  Node(r,s,d,ob) => pr.printq
  `node(recb=^(prc r) state=^(prs s) del=^(pr.d d) obuf=[^(pr.lis ob prc)]) `
end
end
