structure pr = struct
local
  open TextIO
  val out = ref stdOut
  fun print_out s = output(!out,s)
  fun print_quote (QUOTE s) = print_out s
    | print_quote (ANTIQUOTE s) = print_out s
in

fun lis l f =   (* print list to buffer *)
  let
    val buf = ref ""
    val rec list2s = fn
      [] => ""
    | [x] => (!buf)^(f x)
    | (x::xs) => (buf := (!buf)^(f x)^","; list2s xs)
  in
    list2s l
  end

fun vec v f =   (* print vector to buffer *)
  let
    val buf = ref ""
    val len = Vector.length v
    val rec vec2s = fn
      0 => ""
    | 1 => (!buf)^(f (Vector.sub(v,len-1)))
    | n => (buf := (!buf)^(f (Vector.sub(v,len-n)))^","; vec2s(n-1))
  in
    vec2s len
  end

fun printl lis =  (* print string list to stdOut *)
  let 
  fun f [] = ()
    | f (s::ss) = (print_out s; f ss)
  in
    out:=stdOut;
    f lis
  end

fun fprintl fd lis =  (* print string list to file *)
  let 
  fun f [] = ()
    | f (s::ss) = (print_out s; f ss)
  in
    out:=fd;
    f lis
  end

fun sprintl str =  (* print string list to buffer *)
  let
    val buf = ref ""
    fun f [] = !buf
      | f (s::ss) = (buf := (!buf) ^ s; f ss)
  in
    f str
  end

fun d v = Int.toString v         (* string from int *)

fun b b = if b then "T" else "F" (* string from bool *)

fun printq l =       (* print quoted string to stdOut *)
  let 
  fun f [] = ()
    | f (s::ss) = (print_quote s; f ss)
  in
    out:=stdOut;
    f l
  end

fun fprintq fd l =   (* print quoted string to file *)
  let 
  fun f [] = ()
    | f (s::ss) = (print_quote s; f ss)
  in
    out:=fd;
    f l
  end

fun sprintq str =    (* print quoted string to buffer *)
  let
    fun get_quote (QUOTE s) = s
      | get_quote (ANTIQUOTE s) = s
    val buf = ref ""
    fun f [] = !buf
      | f (s::ss) = (buf := (!buf) ^ (get_quote s); f ss)
  in
    f str
  end

fun err qs =       (* print quoted string to stdOut and exit *)
  raise Fail(sprintq qs)

end
end
